# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
"""Tests for tensorflow_datasets.core.deprecated.text.text_encoder."""
from __future__ import unicode_literals

import os

from absl.testing import parameterized
import numpy as np
from tensorflow_datasets import testing
from tensorflow_datasets.core.deprecated.text import text_encoder

ZH_HELLO = '你好 '
EN_HELLO = 'hello '


class TextEncoderTest(testing.TestCase):

  def test_pad_decr(self):
    self.assertEqual([2, 1, 0], text_encoder.pad_decr([3, 2, 1]))
    self.assertEqual([2, 1, 0], text_encoder.pad_decr([3, 2, 1, 0, 0, 0]))
    self.assertEqual([-1, 2, 1, 0], text_encoder.pad_decr([0, 3, 2, 1, 0, 0]))
    self.assertEqual([], text_encoder.pad_decr([]))
    self.assertEqual([], text_encoder.pad_decr(np.array([])))

  def test_pad_incr(self):
    self.assertEqual([4, 3, 2], text_encoder.pad_incr([3, 2, 1]))
    self.assertEqual([4, 3, 2, 1], text_encoder.pad_incr([3, 2, 1, 0]))

  def test_is_mixed_alphanum(self):
    self.assertFalse(text_encoder.is_mixed_alphanum('hi'))
    self.assertFalse(text_encoder.is_mixed_alphanum(ZH_HELLO[:-1]))
    self.assertTrue(text_encoder.is_mixed_alphanum('hi.'))
    self.assertTrue(text_encoder.is_mixed_alphanum('hi.bye'))
    self.assertTrue(text_encoder.is_mixed_alphanum('hi '))
    self.assertTrue(text_encoder.is_mixed_alphanum(ZH_HELLO))


class ByteTextEncoderTest(parameterized.TestCase, testing.TestCase):
  # Incremented for pad
  ZH_HELLO_IDS = [i + 1 for i in [228, 189, 160, 229, 165, 189, 32]]
  EN_HELLO_IDS = [i + 1 for i in [104, 101, 108, 108, 111, 32]]

  def test_encode_decode(self):
    encoder = text_encoder.ByteTextEncoder()
    self.assertEqual(self.ZH_HELLO_IDS, encoder.encode(ZH_HELLO))
    self.assertEqual(self.EN_HELLO_IDS, encoder.encode(EN_HELLO))
    self.assertEqual(self.EN_HELLO_IDS, encoder.encode('hello '))
    self.assertEqual(EN_HELLO, encoder.decode(self.EN_HELLO_IDS))
    self.assertEqual(ZH_HELLO, encoder.decode(self.ZH_HELLO_IDS))
    self.assertEqual(text_encoder.NUM_BYTES + 1, encoder.vocab_size)

  def test_additional_tokens(self):
    # One with non-alphanumeric chars, one uppercase, one lowercase
    additional_tokens = ['<EOS>', 'FOO', 'bar']
    encoder = text_encoder.ByteTextEncoder(additional_tokens=additional_tokens)
    # Without additional tokens
    hello_ids = [i + len(additional_tokens) for i in self.ZH_HELLO_IDS]
    self.assertEqual(hello_ids, encoder.encode(ZH_HELLO))
    # With additional tokens
    text_additional = '%s %s%s%s' % (
        additional_tokens[0],
        ZH_HELLO,
        additional_tokens[1],
        additional_tokens[2],
    )
    expected_ids = [1, 32 + 1 + len(additional_tokens)] + hello_ids + [2, 3]
    self.assertEqual(expected_ids, encoder.encode(text_additional))
    self.assertEqual(text_additional, encoder.decode(expected_ids))
    self.assertEqual(
        text_encoder.NUM_BYTES + 1 + len(additional_tokens), encoder.vocab_size
    )

  @parameterized.parameters(
      (['<EOS>', 'FOO', 'bar'],),
      ([],),
  )
  def test_file_backed(self, additional_tokens):
    encoder = text_encoder.ByteTextEncoder(additional_tokens=additional_tokens)
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      vocab_fname = os.path.join(tmp_dir, 'vocab')
      encoder.save_to_file(vocab_fname)

      file_backed_encoder = text_encoder.ByteTextEncoder.load_from_file(
          vocab_fname
      )
      self.assertEqual(encoder.vocab_size, file_backed_encoder.vocab_size)
      self.assertEqual(
          encoder.additional_tokens, file_backed_encoder.additional_tokens
      )


class TokenTextEncoderTest(testing.TestCase):

  def test_encode_decode(self):
    encoder = text_encoder.TokenTextEncoder(vocab_list=['hi', 'bye', ZH_HELLO])
    ids = [i + 1 for i in [0, 1, 2, 0]]
    self.assertEqual(ids, encoder.encode('hi  bye %s hi' % ZH_HELLO))
    self.assertEqual('hi bye %shi' % ZH_HELLO, encoder.decode(ids))

  def test_oov(self):
    encoder = text_encoder.TokenTextEncoder(
        vocab_list=['hi', 'bye', ZH_HELLO], oov_buckets=1, oov_token='UNK'
    )
    ids = [i + 1 for i in [0, 3, 3, 1]]
    self.assertEqual(ids, encoder.encode('hi boo foo bye'))
    self.assertEqual('hi UNK UNK bye', encoder.decode(ids))
    self.assertEqual(5, encoder.vocab_size)

  def test_multiple_oov(self):
    encoder = text_encoder.TokenTextEncoder(
        vocab_list=['hi', 'bye', ZH_HELLO], oov_buckets=2, oov_token='UNK'
    )
    encoded = encoder.encode('hi boo zoo too foo bye')
    self.assertEqual(1, encoded[0])
    self.assertEqual(2, encoded[-1])
    self.assertIn(4, encoded)
    self.assertIn(5, encoded)
    self.assertEqual(6, encoder.vocab_size)
    self.assertEqual('hi UNK UNK bye', encoder.decode([1, 4, 5, 2]))

  def test_tokenization(self):
    encoder = text_encoder.TokenTextEncoder(vocab_list=['hi', 'bye', ZH_HELLO])
    text = 'hi<<>><<>foo!^* bar && bye (%s hi)' % ZH_HELLO
    self.assertEqual(
        ['hi', 'foo', 'bar', 'bye', ZH_HELLO.strip(), 'hi'],
        text_encoder.Tokenizer().tokenize(text),
    )
    self.assertEqual([i + 1 for i in [0, 3, 3, 1, 2, 0]], encoder.encode(text))

  def test_file_backed(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      vocab_fname = os.path.join(tmp_dir, 'vocab')
      encoder = text_encoder.TokenTextEncoder(
          vocab_list=['hi', 'bye', ZH_HELLO]
      )
      encoder.save_to_file(vocab_fname)
      file_backed_encoder = text_encoder.TokenTextEncoder.load_from_file(
          vocab_fname
      )
      self.assertEqual(encoder.tokens, file_backed_encoder.tokens)

  def test_file_backed_with_args(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Set all the args to non-default values, including Tokenizer
      tokenizer = text_encoder.Tokenizer(
          reserved_tokens=['<FOOBAR>'], alphanum_only=False
      )
      encoder = text_encoder.TokenTextEncoder(
          vocab_list=['hi', 'bye', ZH_HELLO],
          lowercase=True,
          oov_buckets=2,
          oov_token='ZOO',
          tokenizer=tokenizer,
      )

      vocab_fname = os.path.join(tmp_dir, 'vocab')
      encoder.save_to_file(vocab_fname)

      file_backed_encoder = text_encoder.TokenTextEncoder.load_from_file(
          vocab_fname
      )
      self.assertEqual(encoder.tokens, file_backed_encoder.tokens)
      self.assertEqual(encoder.vocab_size, file_backed_encoder.vocab_size)
      self.assertEqual(encoder.lowercase, file_backed_encoder.lowercase)
      self.assertEqual(encoder.oov_token, file_backed_encoder.oov_token)
      self.assertEqual(
          encoder.tokenizer.alphanum_only,
          file_backed_encoder.tokenizer.alphanum_only,
      )
      self.assertEqual(
          encoder.tokenizer.reserved_tokens,
          file_backed_encoder.tokenizer.reserved_tokens,
      )

  def test_mixedalphanum_tokens(self):
    mixed_tokens = ['<EOS>', 'zoo!', '!foo']
    vocab_list = mixed_tokens + ['hi', 'bye', ZH_HELLO]

    encoder = text_encoder.TokenTextEncoder(vocab_list=vocab_list)

    # No mixed tokens
    text = 'hi<<>><<>foo!^* bar && bye (%s hi)' % ZH_HELLO
    # hi=3, foo=OOV, bar=OOV, bye=4, ZH_HELLO=5, hi=3
    text_ids = [i + 1 for i in [3, 6, 6, 4, 5, 3]]
    self.assertEqual(text_ids, encoder.encode(text))

    # With mixed tokens
    text = 'hi<<>><<>foo!<EOS>^* barzoo! FOO && bye (%s hi)' % ZH_HELLO
    # hi=3, foo=OOV, <EOS>=0, bar=OOV, zoo!=1, FOO=OOV, bye=4, ZH_HELLO=5, hi=3
    text_ids = [i + 1 for i in [3, 6, 0, 6, 1, 6, 4, 5, 3]]
    self.assertEqual(text_ids, encoder.encode(text))

  def test_lowercase(self):
    mixed_tokens = ['<EOS>', 'zoo!']
    vocab_list = mixed_tokens + ['hi', 'bye', ZH_HELLO]
    encoder = text_encoder.TokenTextEncoder(
        vocab_list=vocab_list, lowercase=True
    )
    # No mixed tokens
    self.assertEqual([3, 4, 3], encoder.encode('hi byE HI!'))
    # With mixed tokens
    self.assertEqual([3, 1, 4, 2, 3], encoder.encode('hi<EOS>byE Zoo! HI!'))

  def test_with_tokenizer(self):
    class DummyTokenizer(object):

      def tokenize(self, s):
        del s
        return ['hi', 'bye']

    tokenizer = DummyTokenizer()
    vocab_list = ['hi', 'bye', ZH_HELLO]
    encoder = text_encoder.TokenTextEncoder(
        vocab_list=vocab_list, tokenizer=tokenizer
    )
    # Ensure it uses the passed tokenizer and not the default
    self.assertEqual([1, 2], encoder.encode('zoo foo'))


class TokenizeTest(parameterized.TestCase, testing.TestCase):

  def test_default(self):
    text = 'hi<<>><<>foo!^* bar &&  bye (%s hi)' % ZH_HELLO
    self.assertEqual(
        ['hi', 'foo', 'bar', 'bye', ZH_HELLO.strip(), 'hi'],
        text_encoder.Tokenizer().tokenize(text),
    )

  def test_with_nonalphanum(self):
    text = 'hi world<<>><<>foo!^* bar &&  bye (%s hi)' % ZH_HELLO
    tokens = [
        'hi',
        ' ',
        'world',
        '<<>><<>',
        'foo',
        '!^* ',
        'bar',
        ' &&  ',
        'bye',
        ' (',
        ZH_HELLO.strip(),
        '  ',
        'hi',
        ')',
    ]
    tokenizer = text_encoder.Tokenizer(alphanum_only=False)
    self.assertEqual(tokens, tokenizer.tokenize(text))
    self.assertEqual(text, tokenizer.join(tokenizer.tokenize(text)))

  @parameterized.parameters(
      # Single Space at at beginning
      (' hello world', [' ', 'hello', ' ', 'world']),
      (' !hello world', [' !', 'hello', ' ', 'world']),
      # Single space at end
      ('hello world ', ['hello', ' ', 'world', ' ']),
      ('hello world! ', ['hello', ' ', 'world', '! ']),
      # Double space at beginning
      ('  hello world', ['  ', 'hello', ' ', 'world']),
      ('  !hello world', ['  !', 'hello', ' ', 'world']),
      # Double space in middle
      ('hello  world', ['hello', '  ', 'world']),
      ('hello!  world', ['hello', '!  ', 'world']),
      ('hello  !world', ['hello', '  !', 'world']),
  )
  def test_whitespace(self, s, exp):
    tokenizer = text_encoder.Tokenizer(alphanum_only=False)
    self.assertEqual(exp, tokenizer.tokenize(s))
    self.assertEqual(s, tokenizer.join(tokenizer.tokenize(s)))

  def test_reserved_tokens(self):
    text = 'hello worldbar bar foozoo zoo FOO<EOS>'
    tokens = [
        'hello',
        ' ',
        'world',
        'bar',
        ' ',
        'bar',
        ' ',
        'foozoo',
        ' ',
        'zoo',
        ' ',
        'FOO',
        '<EOS>',
    ]
    tokenizer = text_encoder.Tokenizer(
        alphanum_only=False, reserved_tokens=['<EOS>', 'FOO', 'bar']
    )
    self.assertEqual(tokens, tokenizer.tokenize(text))
    self.assertEqual(text, tokenizer.join(tokenizer.tokenize(text)))

  def test_reserved_tokens_with_regex_chars(self):
    text = r'hello worldba\)r bar foozoo zoo FOO|<EOS>'
    tokens = [
        'hello',
        ' ',
        'world',
        r'ba\)r',
        ' ',
        'bar',
        ' ',
        'foozoo',
        ' ',
        'zoo',
        ' ',
        'FOO|',
        '<EOS>',
    ]
    tokenizer = text_encoder.Tokenizer(
        alphanum_only=False, reserved_tokens=['<EOS>', 'FOO|', r'ba\)r']
    )
    self.assertEqual(tokens, tokenizer.tokenize(text))
    self.assertEqual(text, tokenizer.join(tokenizer.tokenize(text)))


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset builder for Huggingface datasets.

Instead of changing the Huggingface dataset builder code to directly construct a
TFDS dataset, here we first download and prepare a Huggingface dataset and use
the resulting dataset to create a new TFDS dataset. This is to support
Huggingface community datasets that are hosted on external repositories.

Furthermore, this also enables creating datasets based on datasets in
Huggingface.
"""
from __future__ import annotations

import datetime
import functools
import io
import itertools
import multiprocessing
import os
from typing import Any, Dict, Mapping, Optional, Tuple, Type, Union

from absl import logging
from etils import epath
import numpy as np
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import download
from tensorflow_datasets.core import features as feature_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import dtype_utils
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_IMAGE_ENCODING_FORMAT = "png"
_EMPTY_SPLIT_WARNING_MSG = "%s split doesn't have any examples"


def _convert_to_np_dtype(dtype: str) -> Type[np.generic]:
  """Returns the `np.dtype` scalar feature."""
  str2val = {
      "bool": np.bool_,
      "bool_": np.bool_,
      "float": np.float32,
      "double": np.float64,
      "large_string": np.object_,
      "utf8": np.object_,
      "string": np.object_,
  }
  if dtype in str2val:
    return str2val[dtype]
  elif hasattr(np, dtype):
    return getattr(np, dtype)
  elif dtype.startswith("timestamp"):
    # Timestamps are converted to seconds since UNIX epoch.
    return np.int64
  elif hasattr(tf.dtypes, dtype):
    return getattr(tf.dtypes, dtype)
  else:
    raise ValueError(
        f"Unrecognized type {dtype}. Please open an issue if you think "
        "this is a bug."
    )


def extract_features(hf_features) -> feature_lib.FeatureConnector:
  """Converts Huggingface feature spec to TFDS feature spec."""
  hf_datasets = lazy_imports_lib.lazy_imports.datasets
  if isinstance(hf_features, (hf_datasets.Features, dict)):
    return feature_lib.FeaturesDict(
        {
            name: extract_features(hf_inner_feature)
            for name, hf_inner_feature in hf_features.items()
        }
    )
  if isinstance(hf_features, hf_datasets.Sequence):
    return feature_lib.Sequence(feature=extract_features(hf_features.feature))
  if isinstance(hf_features, list):
    if len(hf_features) != 1:
      raise ValueError(f"List {hf_features} should have a length of 1.")
    return feature_lib.Sequence(feature=extract_features(hf_features[0]))
  if isinstance(hf_features, hf_datasets.Value):
    return feature_lib.Scalar(dtype=_convert_to_np_dtype(hf_features.dtype))
  if isinstance(hf_features, hf_datasets.ClassLabel):
    if hf_features.names:
      return feature_lib.ClassLabel(names=hf_features.names)
    if hf_features.names_file:
      return feature_lib.ClassLabel(names_file=hf_features.names_file)
    if hf_features.num_classes:
      return feature_lib.ClassLabel(num_classes=hf_features.num_classes)
  if isinstance(hf_features, hf_datasets.Translation):
    return feature_lib.Translation(
        languages=hf_features.languages,
    )
  if isinstance(hf_features, hf_datasets.TranslationVariableLanguages):
    return feature_lib.TranslationVariableLanguages(
        languages=hf_features.languages,
    )
  if isinstance(hf_features, hf_datasets.Image):
    return feature_lib.Image(encoding_format=_IMAGE_ENCODING_FORMAT)
  if isinstance(hf_features, hf_datasets.Audio):
    return feature_lib.Audio(sample_rate=hf_features.sampling_rate)
  raise ValueError(f"Type {type(hf_features)} is not supported.")


def from_hf_to_tfds(hf_name: str) -> str:
  """Converts Huggingface dataset name to a TFDS compatible name.

  Huggingface dataset names can contain characters that are not supported in
  TFDS. For example, in Huggingface a dataset name like `a/b` is supported,
  while in TFDS `b` would be parsed as the config.

  Examples:
  - `hf_name='codeparrot/github-code'` becomes `codeparrot__github_code`.

  Arguments:
    hf_name: the dataset name in Huggingface.

  Returns:
    the TFDS compatible dataset name.
  """
  return hf_name.replace("-", "_").replace(".", "_").replace("/", "__").lower()


def _from_tfds_to_hf(tfds_name: str) -> str:
  """Finds the original HF repo ID.

  As TFDS doesn't support case-sensitive names, we list all HF datasets and pick
  the dataset that has a case-insensitive match.

  Args:
    tfds_name: the dataset name in TFDS.

  Returns:
    the HF dataset name.

  Raises:
    Exception: if the name doesn't correspond to any existing dataset.
  """
  hf_datasets = lazy_imports_lib.lazy_imports.datasets
  hf_names = hf_datasets.list_datasets()
  for hf_name in hf_names:
    if from_hf_to_tfds(hf_name) == tfds_name.lower():
      return hf_name
  raise registered.DatasetNotFoundError(
      f'"{tfds_name}" is not listed in Hugging Face datasets.'
  )


def convert_config_name(hf_config: Optional[str]) -> Optional[str]:
  if hf_config is None:
    return hf_config
  return hf_config.lower()


def _convert_value(hf_value: Any, feature: feature_lib.FeatureConnector) -> Any:
  """Converts a Huggingface value to a TFDS compatible value."""
  # See the docstring of _default_value for explanations.
  if hf_value is None:
    return _default_value(feature)
  if isinstance(hf_value, datetime.datetime):
    return int(hf_value.timestamp())
  elif isinstance(feature, feature_lib.ClassLabel):
    return hf_value
  elif isinstance(feature, feature_lib.Scalar):
    return hf_value
  elif isinstance(feature, feature_lib.Translation):
    if isinstance(hf_value, dict):
      # Replaces `None` values with the default value.
      return {
          key: value if value is not None else _default_value(feature[key])
          for key, value in hf_value.items()
      }
  elif isinstance(feature, feature_lib.FeaturesDict):
    if isinstance(hf_value, dict):
      return {k: _convert_value(v, feature[k]) for k, v in hf_value.items()}
    raise ValueError(f"The feature is {feature}, but the value is: {hf_value}")
  elif isinstance(feature, feature_lib.Sequence):
    if isinstance(hf_value, dict):
      # Should be a dict of lists:
      return {
          k: [_convert_value(el, feature.feature[k]) for el in v]
          for k, v in hf_value.items()
      }
    if isinstance(hf_value, list):
      return [_convert_value(v, feature.feature) for v in hf_value]
    else:
      return [hf_value]
  elif isinstance(feature, feature_lib.Audio):
    assert isinstance(hf_value, dict), f"Audio {hf_value} should be a dict"
    if "array" in hf_value:
      sample_rate = feature.sample_rate
      # Hugging Face uses float, TFDS uses integers.
      return [int(s * sample_rate) for s in hf_value["array"]]
    if "path" in hf_value:
      path = epath.Path(hf_value["path"])
      if path.exists():
        return path
    else:
      raise ValueError(f"{hf_value} is not a valid audio feature.")
  elif isinstance(hf_value, lazy_imports_lib.lazy_imports.PIL_Image.Image):
    buffer = io.BytesIO()
    hf_value.save(fp=buffer, format=_IMAGE_ENCODING_FORMAT)
    return buffer.getvalue()
  elif isinstance(feature, feature_lib.Tensor):
    return hf_value
  raise ValueError(
      f"Type {type(hf_value)} of value {hf_value} "
      f"for feature {type(feature)} is not supported."
  )


def _convert_example(
    index: int,
    example: Mapping[str, Any],
    features: feature_lib.FeaturesDict,
) -> Tuple[int, Mapping[str, Any]]:
  """Converts an example from Huggingface format to TFDS format."""
  converted_example = {
      name: _convert_value(value, features[name])
      for name, value in example.items()
  }
  return index, converted_example


def _extract_supervised_keys(hf_info):
  if hf_info.supervised_keys is not None:
    sk_input = hf_info.supervised_keys.input
    sk_output = hf_info.supervised_keys.output
    if sk_input is not None and sk_output is not None:
      return (sk_input, sk_output)
  return None


def _default_value(
    feature: feature_lib.FeatureConnector,
) -> Union[bytes, int, bool, float]:
  """Returns the default value for a feature.

  Hugging Face is loose as far as typing is concerned. It accepts None values.
  As long as `tfds.features.Optional` does not exist, we default to a constant
  default value.

  For int and float, we do not return 0 or -1, but rather -inf, as 0 or -1 can
  be contained in the values of the dataset. In practice, you can compare your
  value to:

  ```
  np.iinfo(np.int32).min  # for integers
  np.finfo(np.float32).min  # for floats
  ...
  ```

  Args:
    feature: the TFDS feature from which we want the default value.

  Returns:
    The default value.
  """
  if isinstance(feature, feature_lib.FeaturesDict):
    return {
        name: _default_value(sub_feature)
        for name, sub_feature in feature.items()
    }
  elif isinstance(feature, feature_lib.Sequence):
    return []
  elif dtype_utils.is_string(feature.np_dtype):
    return b""
  elif dtype_utils.is_integer(feature.np_dtype):
    return np.iinfo(feature.np_dtype).min
  elif dtype_utils.is_floating(feature.np_dtype):
    return np.finfo(feature.np_dtype).min
  elif dtype_utils.is_bool(feature.np_dtype):
    return False
  raise ValueError(f"Could not get default value for {feature}")


def _remove_empty_splits(
    splits: Dict[str, split_builder_lib.SplitGenerator]
) -> Dict[str, split_builder_lib.SplitGenerator]:
  """Removes empty splits."""
  non_empty_splits = {}

  for split, examples_iterable in splits.items():
    examples_iterator = iter(examples_iterable)
    # ensure the iterator is not empty
    try:
      first_example = next(examples_iterator)
      non_empty_splits[split] = itertools.chain(
          [first_example], examples_iterator
      )
    except StopIteration:
      logging.warning(_EMPTY_SPLIT_WARNING_MSG, split)

  return non_empty_splits


class HuggingfaceDatasetBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """A TFDS builder for Huggingface datasets.

  If a Huggingface config name is given to this builder, it will construct a
  TFDS BuilderConfig. Note that TFDS has some restrictions on config names such
  as it is not allowed to use the config name `all`. Therefore, `all` is
  converted to `_all`.
  """

  VERSION = utils.Version("1.0.0")  # This will be replaced in __init__.

  def __init__(
      self,
      *,
      file_format: Optional[Union[str, file_adapters.FileFormat]] = None,
      hf_repo_id: str,
      hf_config: Optional[str] = None,
      ignore_verifications: bool = False,
      data_dir: Optional[epath.PathLike] = None,
      hf_hub_token: Optional[str] = None,
      hf_num_proc: Optional[int] = None,
      tfds_num_proc: Optional[int] = None,
      **config_kwargs,
  ):
    self._hf_repo_id = hf_repo_id
    self._hf_config = hf_config
    self.config_kwargs = config_kwargs
    tfds_config = convert_config_name(hf_config)
    hf_datasets = lazy_imports_lib.lazy_imports.datasets
    self._hf_builder = hf_datasets.load_dataset_builder(
        self._hf_repo_id, self._hf_config, **self.config_kwargs
    )
    self._hf_info = self._hf_builder.info
    version = str(self._hf_info.version or self._hf_builder.VERSION or "1.0.0")
    self.VERSION = utils.Version(version)  # pylint: disable=invalid-name
    if self._hf_config:
      self._converted_builder_config = dataset_builder.BuilderConfig(
          name=tfds_config,
          version=self.VERSION,
          description=self._hf_info.description,
      )
    else:
      self._converted_builder_config = None
    self.name = from_hf_to_tfds(hf_repo_id)
    self._hf_hub_token = hf_hub_token
    self._hf_num_proc = hf_num_proc
    self._tfds_num_proc = tfds_num_proc
    self._verification_mode = (
        "all_checks" if ignore_verifications else "no_checks"
    )
    super().__init__(
        file_format=file_format, config=tfds_config, data_dir=data_dir
    )
    if self._hf_config:
      self._builder_config = self._converted_builder_config
    self.generation_errors = []

  @property
  def builder_config(self) -> Optional[Any]:
    return self._converted_builder_config

  def _create_builder_config(
      self, builder_config
  ) -> Optional[dataset_builder.BuilderConfig]:
    return self._converted_builder_config

  def _download_and_prepare_for_hf(self) -> Mapping[str, Any]:
    login_to_hf(self._hf_hub_token)
    self._hf_builder.download_and_prepare(
        num_proc=self._hf_num_proc,
        verification_mode=self._verification_mode,
    )
    return self._hf_builder.as_dataset(
        verification_mode=self._verification_mode,
    )

  def _hf_features(self):
    if self._hf_info.features is not None:
      return self._hf_info.features
    # We need to download and prepare the data to know its features.
    ds = self._download_and_prepare_for_hf()
    for split in ds.values():
      return split.info

  @py_utils.memoize()
  def _info(self) -> dataset_info_lib.DatasetInfo:
    return dataset_info_lib.DatasetInfo(
        builder=self,
        description=self._hf_info.description,
        features=extract_features(self._hf_features()),
        citation=self._hf_info.citation,
        license=self._hf_info.license,
        supervised_keys=_extract_supervised_keys(self._hf_info),
    )

  def _split_generators(
      self, dl_manager: download.DownloadManager
  ) -> Dict[splits_lib.Split, split_builder_lib.SplitGenerator]:
    del dl_manager
    ds = self._download_and_prepare_for_hf()
    splits = {
        split: self._generate_examples(data) for split, data in ds.items()
    }
    return _remove_empty_splits(splits)

  def _generate_examples(self, data) -> split_builder_lib.SplitGenerator:
    dataset_info = self._info()
    if self._tfds_num_proc is None:
      for index, example in enumerate(data):
        yield _convert_example(index, example, dataset_info.features)
    else:
      with multiprocessing.Pool(processes=self._tfds_num_proc) as pool:
        examples = pool.starmap(
            functools.partial(_convert_example, features=dataset_info.features),
            enumerate(data),
        )
        yield from examples


def builder(
    name: str, config: Optional[str] = None, **builder_kwargs
) -> HuggingfaceDatasetBuilder:
  hf_repo_id = _from_tfds_to_hf(name)
  return HuggingfaceDatasetBuilder(
      hf_repo_id=hf_repo_id, hf_config=config, **builder_kwargs
  )


def login_to_hf(hf_hub_token: Optional[str] = None):
  """Logs in to Hugging Face Hub with the token as arg or env variable."""
  hf_hub_token = hf_hub_token or os.environ.get("HUGGING_FACE_HUB_TOKEN")
  if hf_hub_token is not None:
    huggingface_hub = lazy_imports_lib.lazy_imports.huggingface_hub
    huggingface_hub.login(token=hf_hub_token)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for libsvm_ranking_parser."""

from tensorflow_datasets import testing
from tensorflow_datasets.ranking.libsvm_ranking_parser import LibSVMRankingParser
from tensorflow_datasets.ranking.libsvm_ranking_parser import ParserError

# The following is a test dataset that exhibits some of the LibSVM edge cases
# which should be appropriately handled by the parser.
_TEST_DATASET = """
3 qid:1 1:1 2:1 3:0 4:0.2 5:0  # Line-comments should be ignored.
1 qid:1 1:0 2:0 3:1 4:0.1 5:1
0  qid:1  \t 1:0   2:1   3:0   4:0.4   5:0  # Extra spacing in a line is fine.
0 qid:1 1:0 2:0 3:1 4:0.3 5:0
0 qid:2 1:0 2:0 3:1 4:0.2 5:0  # Feature values after comment should be ignored:
1 qid:2 1:1 2:0 3:1 4:0.4 5:0  # 1:10 2:20 3:30 4:40 5:50
# just a comment, this entire line should be ignored by the parser.
# empty lines should also be ignored:

0 qid:2 1:0 2:0 3:1 4:0.1 5:0
0 qid:2 1:0 2:0 3:1 4:0.2 5:0

# feature 6 on the next line will be ignored if it is not specified in the
# feature_names map of the parser (see first unit test):
1 qid:2 1:0 2:0 3:1 4:0.1 5:1 6:2
   # Leading and trailing white space should be ignored:
   2 qid:2 1:1 2:1 3:0 4:0.3 5:0  \t
3 qid:3 1:1         4:0.4 5:1  # missing features should have default value (0)
0 qid:3 1:0 2:1 3:1 4:0.5 5:0
"""


class LibSVMRankingParserTest(testing.TestCase):

  def test_parses_a_libsvm_ranking_dataset(self):
    results = iter(
        LibSVMRankingParser(
            _TEST_DATASET.split("\n"),
            feature_names={
                1: "bm25",
                2: "tfidf",
                3: "querylen",
                4: "doclen",
                5: "qualityscore",
            },
        )
    )

    qid, features = next(results)
    self.assertEqual(qid, "1")
    self.assertEqual(features["query_id"], "1")
    self.assertAllEqual(features["doc_id"], [2, 3, 4, 5])
    self.assertAllEqual(features["label"], [3.0, 1.0, 0.0, 0.0])
    self.assertAllEqual(features["bm25"], [1.0, 0.0, 0.0, 0.0])
    self.assertAllEqual(features["tfidf"], [1.0, 0.0, 1.0, 0.0])
    self.assertAllEqual(features["querylen"], [0.0, 1.0, 0.0, 1.0])
    self.assertAllEqual(features["doclen"], [0.2, 0.1, 0.4, 0.3])
    self.assertAllEqual(features["qualityscore"], [0.0, 1.0, 0.0, 0])

    qid, features = next(results)
    self.assertEqual(qid, "2")
    self.assertEqual(features["query_id"], "2")
    self.assertAllEqual(features["doc_id"], [6, 7, 11, 12, 16, 18])
    self.assertAllEqual(features["label"], [0.0, 1.0, 0.0, 0.0, 1.0, 2.0])
    self.assertAllEqual(features["bm25"], [0.0, 1.0, 0.0, 0.0, 0.0, 1.0])
    self.assertAllEqual(features["tfidf"], [0.0, 0.0, 0.0, 0.0, 0.0, 1.0])
    self.assertAllEqual(features["querylen"], [1.0, 1.0, 1.0, 1.0, 1.0, 0.0])
    self.assertAllEqual(features["doclen"], [0.2, 0.4, 0.1, 0.2, 0.1, 0.3])
    self.assertAllEqual(
        features["qualityscore"], [0.0, 0.0, 0.0, 0.0, 1.0, 0.0]
    )

    qid, features = next(results)
    self.assertEqual(qid, "3")
    self.assertEqual(features["query_id"], "3")
    self.assertAllEqual(features["doc_id"], [19, 20])
    self.assertAllEqual(features["label"], [3.0, 0.0])
    self.assertAllEqual(features["bm25"], [1.0, 0.0])
    self.assertAllEqual(features["tfidf"], [0.0, 1.0])
    self.assertAllEqual(features["querylen"], [0.0, 1.0])
    self.assertAllEqual(features["doclen"], [0.4, 0.5])
    self.assertAllEqual(features["qualityscore"], [1.0, 0.0])

    # Assert that the end of the file has been reached.
    with self.assertRaises(StopIteration):
      next(results)

  def test_combines_features_into_a_single_matrix(self):
    results = iter(
        LibSVMRankingParser(
            _TEST_DATASET.split("\n"),
            feature_names={
                1: "bm25",
                2: "tfidf",
                3: "querylen",
                4: "doclen",
                5: "qualityscore",
            },
            combine_features=True,
        )
    )

    _, features = next(results)
    self.assertAllEqual(
        features["float_features"],
        [
            [1.0, 1.0, 0.0, 0.2, 0.0],
            [0.0, 0.0, 1.0, 0.1, 1.0],
            [0.0, 1.0, 0.0, 0.4, 0.0],
            [0.0, 0.0, 1.0, 0.3, 0.0],
        ],
    )

    _, features = next(results)
    self.assertAllEqual(
        features["float_features"],
        [
            [0.0, 0.0, 1.0, 0.2, 0.0],
            [1.0, 0.0, 1.0, 0.4, 0.0],
            [0.0, 0.0, 1.0, 0.1, 0.0],
            [0.0, 0.0, 1.0, 0.2, 0.0],
            [0.0, 0.0, 1.0, 0.1, 1.0],
            [1.0, 1.0, 0.0, 0.3, 0.0],
        ],
    )

    _, features = next(results)
    self.assertAllEqual(
        features["float_features"],
        [[1.0, 0.0, 0.0, 0.4, 1.0], [0.0, 1.0, 1.0, 0.5, 0.0]],
    )

  def test_raises_error_if_line_only_contains_label(self):
    contents = ["1"]
    with self.assertRaisesRegex(
        ParserError, "could not extract label, qid and features"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_label_is_not_a_number(self):
    contents = ["abc qid:1 2:0.0 3:0.0"]
    with self.assertRaisesRegex(
        ParserError, "label 'abc' could not be converted to a float"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_qid_is_missing(self):
    contents = ["0 2:0.0 3:0.0"]
    with self.assertRaisesRegex(
        ParserError, "line must contain a qid after the relevance label"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_qid_is_malformatted(self):
    contents = ["1 qid: 2:0.0 3:0.0"]
    with self.assertRaisesRegex(ParserError, "qid can not be empty"):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_feature_index_is_malformatted(self):
    contents = ["1 qid:1 2:0.0 _12_:0.0"]
    with self.assertRaisesRegex(
        ParserError, "failed to extract feature index and value from '_12_:0.0'"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_feature_value_is_malformatted(self):
    contents = ["1 qid:1 2:0.0 3:0.00.1 4:10.0"]
    with self.assertRaisesRegex(
        ParserError, "failed to extract feature index and value from '3:0.00.1'"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_if_feature_is_not_separated_by_colon(self):
    contents = ["1 qid:1 2:0.0 3=0.1 4:10.0"]
    with self.assertRaisesRegex(
        ParserError, "failed to extract feature index and value from '3=0.1'"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_with_offending_line_number(self):
    contents = ["1 qid:1 2:0.0", "malformatted line", "1 qid:1 2:0.0"]
    with self.assertRaisesWithPredicateMatch(
        ParserError, lambda error: error.line_number == 2
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass

  def test_raises_error_with_offending_line(self):
    contents = ["1 qid:1 2:1000.0", "2 qid:1 9:9.9", "malformatted line"]
    with self.assertRaisesWithPredicateMatch(
        ParserError, lambda error: error.line == "malformatted line"
    ):
      for _ in LibSVMRankingParser(contents, {}):
        pass


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DomainNet dataset."""
import dataclasses
import os
from typing import Any, Dict, Iterator, Optional, Text, Tuple

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The DomainNet dataset consists of images from six distinct domains, including
photos (real), painting, clipart, quickdraw, infograph and sketch. Per domain
there are 48K - 172K images (600K in total) categorized into 345 classes.

In this TFDS version of DomainNet the cleaned version is used.
"""

_HOMEPAGE = 'http://ai.bu.edu/DomainNet/'
_CITATION = """\
@inproceedings{peng2019moment,
  title={Moment matching for multi-source domain adaptation},
  author={Peng, Xingchao and Bai, Qinxun and Xia, Xide and Huang, Zijun and Saenko, Kate and Wang, Bo},
  booktitle={Proceedings of the IEEE International Conference on Computer Vision},
  pages={1406--1415},
  year={2019}
}
"""

_CLASS_NAMES = [
    'aircraft_carrier',
    'airplane',
    'alarm_clock',
    'ambulance',
    'angel',
    'animal_migration',
    'ant',
    'anvil',
    'apple',
    'arm',
    'asparagus',
    'axe',
    'backpack',
    'banana',
    'bandage',
    'barn',
    'baseball',
    'baseball_bat',
    'basket',
    'basketball',
    'bat',
    'bathtub',
    'beach',
    'bear',
    'beard',
    'bed',
    'bee',
    'belt',
    'bench',
    'bicycle',
    'binoculars',
    'bird',
    'birthday_cake',
    'blackberry',
    'blueberry',
    'book',
    'boomerang',
    'bottlecap',
    'bowtie',
    'bracelet',
    'brain',
    'bread',
    'bridge',
    'broccoli',
    'broom',
    'bucket',
    'bulldozer',
    'bus',
    'bush',
    'butterfly',
    'cactus',
    'cake',
    'calculator',
    'calendar',
    'camel',
    'camera',
    'camouflage',
    'campfire',
    'candle',
    'cannon',
    'canoe',
    'car',
    'carrot',
    'castle',
    'cat',
    'ceiling_fan',
    'cello',
    'cell_phone',
    'chair',
    'chandelier',
    'church',
    'circle',
    'clarinet',
    'clock',
    'cloud',
    'coffee_cup',
    'compass',
    'computer',
    'cookie',
    'cooler',
    'couch',
    'cow',
    'crab',
    'crayon',
    'crocodile',
    'crown',
    'cruise_ship',
    'cup',
    'diamond',
    'dishwasher',
    'diving_board',
    'dog',
    'dolphin',
    'donut',
    'door',
    'dragon',
    'dresser',
    'drill',
    'drums',
    'duck',
    'dumbbell',
    'ear',
    'elbow',
    'elephant',
    'envelope',
    'eraser',
    'eye',
    'eyeglasses',
    'face',
    'fan',
    'feather',
    'fence',
    'finger',
    'fire_hydrant',
    'fireplace',
    'firetruck',
    'fish',
    'flamingo',
    'flashlight',
    'flip_flops',
    'floor_lamp',
    'flower',
    'flying_saucer',
    'foot',
    'fork',
    'frog',
    'frying_pan',
    'garden',
    'garden_hose',
    'giraffe',
    'goatee',
    'golf_club',
    'grapes',
    'grass',
    'guitar',
    'hamburger',
    'hammer',
    'hand',
    'harp',
    'hat',
    'headphones',
    'hedgehog',
    'helicopter',
    'helmet',
    'hexagon',
    'hockey_puck',
    'hockey_stick',
    'horse',
    'hospital',
    'hot_air_balloon',
    'hot_dog',
    'hot_tub',
    'hourglass',
    'house',
    'house_plant',
    'hurricane',
    'ice_cream',
    'jacket',
    'jail',
    'kangaroo',
    'key',
    'keyboard',
    'knee',
    'knife',
    'ladder',
    'lantern',
    'laptop',
    'leaf',
    'leg',
    'light_bulb',
    'lighter',
    'lighthouse',
    'lightning',
    'line',
    'lion',
    'lipstick',
    'lobster',
    'lollipop',
    'mailbox',
    'map',
    'marker',
    'matches',
    'megaphone',
    'mermaid',
    'microphone',
    'microwave',
    'monkey',
    'moon',
    'mosquito',
    'motorbike',
    'mountain',
    'mouse',
    'moustache',
    'mouth',
    'mug',
    'mushroom',
    'nail',
    'necklace',
    'nose',
    'ocean',
    'octagon',
    'octopus',
    'onion',
    'oven',
    'owl',
    'paintbrush',
    'paint_can',
    'palm_tree',
    'panda',
    'pants',
    'paper_clip',
    'parachute',
    'parrot',
    'passport',
    'peanut',
    'pear',
    'peas',
    'pencil',
    'penguin',
    'piano',
    'pickup_truck',
    'picture_frame',
    'pig',
    'pillow',
    'pineapple',
    'pizza',
    'pliers',
    'police_car',
    'pond',
    'pool',
    'popsicle',
    'postcard',
    'potato',
    'power_outlet',
    'purse',
    'rabbit',
    'raccoon',
    'radio',
    'rain',
    'rainbow',
    'rake',
    'remote_control',
    'rhinoceros',
    'rifle',
    'river',
    'roller_coaster',
    'rollerskates',
    'sailboat',
    'sandwich',
    'saw',
    'saxophone',
    'school_bus',
    'scissors',
    'scorpion',
    'screwdriver',
    'sea_turtle',
    'see_saw',
    'shark',
    'sheep',
    'shoe',
    'shorts',
    'shovel',
    'sink',
    'skateboard',
    'skull',
    'skyscraper',
    'sleeping_bag',
    'smiley_face',
    'snail',
    'snake',
    'snorkel',
    'snowflake',
    'snowman',
    'soccer_ball',
    'sock',
    'speedboat',
    'spider',
    'spoon',
    'spreadsheet',
    'square',
    'squiggle',
    'squirrel',
    'stairs',
    'star',
    'steak',
    'stereo',
    'stethoscope',
    'stitches',
    'stop_sign',
    'stove',
    'strawberry',
    'streetlight',
    'string_bean',
    'submarine',
    'suitcase',
    'sun',
    'swan',
    'sweater',
    'swing_set',
    'sword',
    'syringe',
    'table',
    'teapot',
    'teddy-bear',
    'telephone',
    'television',
    'tennis_racquet',
    'tent',
    'The_Eiffel_Tower',
    'The_Great_Wall_of_China',
    'The_Mona_Lisa',
    'tiger',
    'toaster',
    'toe',
    'toilet',
    'tooth',
    'toothbrush',
    'toothpaste',
    'tornado',
    'tractor',
    'traffic_light',
    'train',
    'tree',
    'triangle',
    'trombone',
    'truck',
    'trumpet',
    't-shirt',
    'umbrella',
    'underwear',
    'van',
    'vase',
    'violin',
    'washing_machine',
    'watermelon',
    'waterslide',
    'whale',
    'wheel',
    'windmill',
    'wine_bottle',
    'wine_glass',
    'wristwatch',
    'yoga',
    'zebra',
    'zigzag',
]


@dataclasses.dataclass
class DomainnetConfig(tfds.core.BuilderConfig):
  name: Text
  img_url: Optional[Text] = None


class Domainnet(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for domainnet dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      DomainnetConfig(name='real', img_url='real.zip'),
      DomainnetConfig(name='painting', img_url='groundtruth/painting.zip'),
      DomainnetConfig(name='clipart', img_url='groundtruth/clipart.zip'),
      DomainnetConfig(name='quickdraw', img_url='quickdraw.zip'),
      DomainnetConfig(name='infograph', img_url='infograph.zip'),
      DomainnetConfig(name='sketch', img_url='sketch.zip'),
  ]

  _BASE_URL = 'http://csr.bu.edu/ftp/visda/2019/multi-source'

  img_path: Any
  splits: Dict[str, Any]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        citation=_CITATION,
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(shape=(None, None, 3)),
            'label': tfds.features.ClassLabel(names=_CLASS_NAMES),
        }),
        homepage=_HOMEPAGE,
        supervised_keys=('image', 'label'),
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    img_download_url = f'{self._BASE_URL}/{self.builder_config.img_url}'
    self.img_path = dl_manager.download_and_extract(img_download_url)

    domain = self.builder_config.name
    train_split_url = f'{self._BASE_URL}/domainnet/txt/{domain}_train.txt'
    test_split_url = f'{self._BASE_URL}/domainnet/txt/{domain}_test.txt'

    self.splits = dl_manager.download(
        {'train': train_split_url, 'test': test_split_url}
    )

    return {
        'train': self._generate_examples(split='train'),
        'test': self._generate_examples(split='test'),
    }

  def _generate_examples(
      self, split: Text = 'train'
  ) -> Iterator[Tuple[Text, Dict[Text, Any]]]:
    """Generator of examples for each split."""
    with tf.io.gfile.GFile(self.splits[split]) as split_file:  # pytype: disable=attribute-error  # gen-stub-imports
      for i, img_class_line in enumerate(split_file.read().split('\n')):
        if not img_class_line:
          continue
        key = f'{self.builder_config.name}_{split}_{i:08d}'

        example_path, example_class = img_class_line.split(' ')
        example_fullpath = os.path.join(self.img_path, example_path)  # pytype: disable=attribute-error  # gen-stub-imports

        yield key, {'image': example_fullpath, 'label': int(example_class)}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper around tqdm."""

import contextlib
import os

from tqdm import auto as tqdm_lib


class TqdmStream:
  """File-object-like abstraction which wrap`tqdm.write`.

  By default using `logging.info` inside a `tqdm` scope creates visual
  artifacts. This simple wrapper uses `tqdm.write` instead.

  Usage:

  ```python
  logger = logging.getLogger()
  logger.addHandler(logging.StreamHandler(TqdmStream()))

  for _ in tqdm.tqdm(range(10)):
    logger.info('No visual artifacts')
  ```
  """

  def write(self, x):
    tqdm_lib.tqdm.write(x, end='')

  def flush(self):
    pass

  def close(self):
    pass


class EmptyTqdm(object):
  """Dummy tqdm which doesn't do anything."""

  def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
    self._iterator = args[0] if args else None

  def __iter__(self):
    return iter(self._iterator)

  def __getattr__(self, _):
    """Return empty function."""

    def empty_fn(*args, **kwargs):  # pylint: disable=unused-argument
      return

    return empty_fn

  def __enter__(self):
    return self

  def __exit__(self, type_, value, traceback):
    return


_active = True
# Disable progression bar when TFDS is executed inside TF kokoro documentation
# infrastructure. Otherwise it creates visual artifacts in the notebook output
# of the documentation pages.
if 'TF_DOCS_INFRA_KOKORO' in os.environ:
  _active = False


def tqdm(*args, **kwargs):
  if _active:
    return tqdm_lib.tqdm(*args, **kwargs)
  else:
    return EmptyTqdm(*args, **kwargs)


def async_tqdm(*args, **kwargs):
  if _active:
    return _async_tqdm(*args, **kwargs)
  else:
    return EmptyTqdm(*args, **kwargs)


def display_progress_bar(enable: bool) -> None:
  """Controls whether Tqdm progress bar is enabled/disabled.

  Usage:

  ```
  tfds.display_progress_bar(enable=True)
  ```

  Args:
    enable: whether to display the progress bar.
  """
  # Replace tqdm
  global _active
  _active = enable


def disable_progress_bar():
  """Disables Tqdm progress bar.

  Usage:

  ```
  tfds.disable_progress_bar()
  ```
  """
  # Replace tqdm
  global _active
  _active = False


def enable_progress_bar():
  """Enables Tqdm progress bar.

  Usage:

  ```
  tfds.enable_progress_bar()
  ```
  """
  # Replace tqdm
  global _active
  _active = True


@contextlib.contextmanager
def _async_tqdm(*args, **kwargs):
  """Wrapper around Tqdm which can be updated in threads.

  Usage:

  ```
  with utils.async_tqdm(...) as pbar:
    # pbar can then be modified inside a thread
    # pbar.update_total(3)
    # pbar.update()
  ```

  Args:
    *args: args of tqdm
    **kwargs: kwargs of tqdm

  Yields:
    pbar: Async pbar which can be shared between threads.
  """
  with tqdm_lib.tqdm(*args, **kwargs) as pbar:
    pbar = _TqdmPbarAsync(pbar)
    yield pbar
    pbar.clear()  # pop pbar from the active list of pbar


class _TqdmPbarAsync(EmptyTqdm):
  """Wrapper around Tqdm pbar which be shared between thread."""

  _tqdm_bars = []

  def __init__(self, pbar):
    super().__init__()
    self._lock = tqdm_lib.tqdm.get_lock()
    self._pbar = pbar
    self._tqdm_bars.append(pbar)

  def update_total(self, n=1):
    """Increment total pbar value."""
    with self._lock:
      self._pbar.total += n
      self.refresh()

  def update(self, n=1):
    """Increment current value."""
    with self._lock:
      self._pbar.update(n)
      self.refresh()

  def refresh(self):
    """Refresh all."""
    for pbar in self._tqdm_bars:
      pbar.refresh()

  def clear(self):
    """Remove the tqdm pbar from the update."""
    self._tqdm_bars.pop()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mujoco_build_configs."""

import tensorflow as tf
from tensorflow_datasets.d4rl import dataset_builder


def _mujoco_replay_datasets():
  """Set of Mujoco datasets with replay."""
  return {
      'v1-medium-replay',
      'v1-full-replay',
      'v2-medium-replay',
      'v2-full-replay',
  }


def _mujoco_v1_replay_datasets():
  """Set of v1 Mujoco datasets with replay."""
  return {
      'v1-medium-replay',
      'v1-full-replay',
  }


def _mujoco_full_metadata_datasets():
  """Set of Mujoco datasets that contain all of the metadata fields."""
  return {'v1-expert', 'v2-expert', 'v1-medium', 'v2-medium'}


class MujocoDatasetTest(tf.test.TestCase):

  def test_builder_config_step_metadata(self):
    for config in dataset_builder.MUJOCO_BUILDER_CONFIGS:
      if config.name in _mujoco_v1_replay_datasets():
        self.assertEqual(config.float_type, tf.float64)
      else:
        self.assertEqual(config.float_type, tf.float32)

  def test_builder_config_float_type(self):
    for config in dataset_builder.MUJOCO_BUILDER_CONFIGS:
      if config.dataset_dir != 'gym_mujoco':
        self.assertTrue(
            config.step_metadata_keys,
            f'Config {config.name} should contain step metadata.',
        )
      else:
        self.assertFalse(
            config.step_metadata_keys,
            f'Config {config.name} should NOT contain step metadata.',
        )

  def test_builder_config_episode_metadata(self):
    for config in dataset_builder.MUJOCO_BUILDER_CONFIGS:
      if config.name in _mujoco_replay_datasets():
        self.assertTrue(
            config.episode_metadata_keys,
            f'Config {config.name} should contain episode metadata.',
        )
        self.assertFalse(
            config.has_policy_metadata,
            f'Config {config.name} should NOT contain policy metadata.',
        )
        self.assertFalse(
            config.has_policy_last_fc_log_std,
            (
                f'Config {config.name} should NOT contain policy values for'
                ' last_fc_log_std.'
            ),
        )
      elif config.name in _mujoco_full_metadata_datasets():
        self.assertTrue(
            config.episode_metadata_keys,
            f'Config {config.name} should contain episode metadata.',
        )
        self.assertTrue(
            config.has_policy_metadata,
            f'Config {config.name} should contain policy metadata.',
        )
        self.assertTrue(
            config.has_policy_last_fc_log_std,
            (
                f'Config {config.name} should contain policy values for'
                ' last_fc_log_std.'
            ),
        )
      else:
        self.assertFalse(
            config.episode_metadata_keys,
            f'Config {config.name} should NOT contain episode metadata.',
        )
        self.assertFalse(
            config.has_policy_metadata,
            f'Config {config.name} should NOT contain policy metadata.',
        )
        self.assertFalse(
            config.has_policy_last_fc_log_std,
            (
                f'Config {config.name} should NOT contain policy values for'
                ' last_fc_log_std.'
            ),
        )


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for view_builder."""
import functools

import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets.core.dataset_builders import view_builder


def add_number(number: int, increment: int) -> int:
  return number + increment


def is_even(number: int) -> bool:
  return number % 2 == 0


_MNIST_TRANSFORMATIONS = [
    tfds.transform.remove_feature(feature_name="image"),
    tfds.transform.apply_filter(fn=is_even, input_feature="label"),
    tfds.transform.apply_fn(
        fn=functools.partial(add_number, increment=10),
        input_feature="label",
        output_feature="label_plus_10",
    ),
    tfds.transform.apply_fn(
        fn=functools.partial(add_number, increment=2),
        input_feature="label",
        output_feature="label_plus_2",
    ),
]

_TRANSFORMED_MNIST_FEATURES = tfds.features.FeaturesDict({
    "label": tfds.features.ClassLabel(num_classes=10),
    "label_plus_10": tfds.features.Scalar(dtype=tf.int64),
    "label_plus_2": tfds.features.Scalar(dtype=tf.int64),
})


def add_number_map_fn(
    dataset: tf.data.Dataset,
    increment: int,
    input_name: str,
    output_name: str,
) -> tf.data.Dataset:
  def f(ex):
    ex[output_name] = ex[input_name] + increment
    return ex

  return dataset.map(f)


def remove_feature_map_fn(
    dataset: tf.data.Dataset,
    feature_name: str,
) -> tf.data.Dataset:
  def f(ex):
    del ex[feature_name]
    return ex

  return dataset.map(f)


_MNIST_DATASET_TRANSFORMATIONS = [
    functools.partial(
        add_number_map_fn,
        increment=10,
        input_name="label",
        output_name="label_plus_10",
    ),
    functools.partial(
        add_number_map_fn,
        increment=2,
        input_name="label",
        output_name="label_plus_2",
    ),
    functools.partial(remove_feature_map_fn, feature_name="image"),
]


class DummyMnistViewWithoutConfigs(view_builder.ViewBuilder):
  VERSION = "1.0.0"
  INPUT_DATASET = "dummy_mnist"
  EX_TRANSFORMATIONS = _MNIST_TRANSFORMATIONS

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        features=_TRANSFORMED_MNIST_FEATURES,
        description="A different view on Mnist.",
    )


class DummyMnistViewWithConfigs(view_builder.ViewBuilder):
  VERSION = "1.0.0"
  BUILDER_CONFIGS = [
      view_builder.ViewConfig(
          name="add_stuff",
          input_dataset="dummy_mnist",
          ex_transformations=_MNIST_TRANSFORMATIONS,
      )
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        features=_TRANSFORMED_MNIST_FEATURES,
        description="A different view on Mnist.",
    )


class BeamDummyMnistViewWithConfigs(view_builder.ViewBuilder):
  VERSION = "1.0.0"
  BUILDER_CONFIGS = [
      view_builder.ViewConfig(
          name="add_stuff",
          input_dataset="dummy_mnist",
          ex_transformations=_MNIST_TRANSFORMATIONS,
      )
  ]
  USE_BEAM = True

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        features=_TRANSFORMED_MNIST_FEATURES,
        description="A different view on Mnist.",
    )


class DummyMnistViewDatasetTransform(view_builder.ViewBuilder):
  VERSION = "1.0.0"
  BUILDER_CONFIGS = [
      view_builder.ViewConfig(
          name="add_stuff",
          input_dataset="dummy_mnist",
          ds_transformations=_MNIST_DATASET_TRANSFORMATIONS,
      )
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        features=_TRANSFORMED_MNIST_FEATURES,
        description="A different view on Mnist.",
    )


def test_view_builder_with_configs_load():
  with tfds.testing.tmp_dir() as data_dir:
    tfds.testing.DummyMnist(data_dir=data_dir).download_and_prepare()

    ds_train = tfds.load(
        "dummy_mnist_view_with_configs", split="train", data_dir=data_dir
    )
    assert len(list(ds_train)) == 10
    for example in ds_train:
      assert example["label"] + 10 == example["label_plus_10"]
      assert example["label"] + 2 == example["label_plus_2"]


def test_beam_view_builder_with_configs_load():
  with tfds.testing.tmp_dir() as data_dir:
    tfds.testing.DummyMnist(data_dir=data_dir).download_and_prepare()

    ds_train = tfds.load(
        "beam_dummy_mnist_view_with_configs", split="train", data_dir=data_dir
    )
    assert len(list(ds_train)) == 10
    for example in ds_train:
      assert example["label"] + 10 == example["label_plus_10"]
      assert example["label"] + 2 == example["label_plus_2"]


def test_view_builder_without_configs_load():
  with tfds.testing.tmp_dir() as data_dir:
    tfds.testing.DummyMnist(data_dir=data_dir).download_and_prepare()

    ds_train = tfds.load(
        "dummy_mnist_view_without_configs", split="train", data_dir=data_dir
    )
    assert len(list(ds_train)) == 10
    for example in ds_train:
      assert example["label"] + 10 == example["label_plus_10"]
      assert example["label"] + 2 == example["label_plus_2"]


def test_view_builder_tf_dataset_with_configs_load():
  with tfds.testing.tmp_dir() as data_dir:
    tfds.testing.DummyMnist(data_dir=data_dir).download_and_prepare()

    ds_train = tfds.load(
        "dummy_mnist_view_dataset_transform", split="train", data_dir=data_dir
    )
    assert len(list(ds_train)) == 20
    for example in ds_train:
      assert example["label"] + 10 == example["label_plus_10"]
      assert example["label"] + 2 == example["label_plus_2"]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dataset_collection_builder."""
from unittest import mock

from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_collection_builder as dcb
from tensorflow_datasets.core.utils import version as version_lib


@mock.patch.object(dcb, 'get_filepath_in_dataset_folder', autospec=True)
def test_get_file_content_from_dataset_folder(mock_get_path):
  expected = 'Mocked file content'
  mock_path = mock.create_autospec(epath.Path)
  mock_path.read_text.return_value = expected
  mock_get_path.return_value = mock_path
  actual = dcb.get_file_content_from_dataset_folder(
      dataset_class=None, file_name='anything'
  )
  mock_get_path.assert_called_once_with(None, 'anything')
  assert actual == expected


@mock.patch.object(dcb, 'get_filepath_in_dataset_folder', autospec=True)
def test_get_file_content_from_dataset_folder_raise_exception(mock_get_path):
  with pytest.raises(AttributeError):
    mock_get_path.return_value = FileNotFoundError
    dcb.get_file_content_from_dataset_folder(
        dataset_class=None, file_name=None, raise_error_if_fails=True
    )


@mock.patch.object(dcb, 'get_filepath_in_dataset_folder', autospec=True)
def test_get_file_content_from_dataset_folder_return_none(mock_get_path):
  mock_get_path.return_value = FileNotFoundError
  actual = dcb.get_file_content_from_dataset_folder(
      dataset_class=None, file_name=None
  )
  assert actual is None


def test_dataset_collection_info_from_cls():
  expected_description = 'description'
  dummy_dc_info = dcb.DatasetCollectionInfo.from_cls(
      dataset_collection_class=testing.DummyDatasetCollection,
      description=expected_description,
      release_notes={
          '1.0.0': 'Initial release',
      },
  )

  assert dummy_dc_info.name == 'dummy_dataset_collection'
  assert dummy_dc_info.description == expected_description
  assert dummy_dc_info.citation is None
  assert dummy_dc_info.homepage is None


@pytest.fixture(name='dummy_dc', scope='session')
def fixture_dummy_dc():
  return testing.DummyDatasetCollection()


def test_all_versions(dummy_dc):
  assert set(dummy_dc.all_versions) == {
      version_lib.Version('1.0.0'),
      version_lib.Version('1.1.0'),
      version_lib.Version('2.0.0'),
  }


def test_get_latest_version(dummy_dc):
  assert dummy_dc.get_latest_version() == '2.0.0'


def test_get_collection(dummy_dc):
  assert sorted(dummy_dc.get_collection().keys()) == ['a', 'b', 'c']
  assert sorted(dummy_dc.get_collection(version='2.0.0').keys()) == [
      'a',
      'b',
      'c',
  ]
  assert sorted(dummy_dc.get_collection(version='1.0.0').keys()) == ['a', 'b']


def test_get_collection_with_wildcards(dummy_dc):
  assert sorted(dummy_dc.get_collection(version='1.0.*').keys()) == ['a', 'b']
  assert sorted(dummy_dc.get_collection(version='1.*.*').keys()) == ['a', 'c']


def test_get_nonexistent_collection(dummy_dc):
  with pytest.raises(ValueError):
    dummy_dc.get_collection(version='1.2.0')
    dummy_dc.get_collection(version='1.*.2')



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Conversational Question Answering Challenge."""

from __future__ import annotations

import json

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@misc{reddy2018coqa,
    title={CoQA: A Conversational Question Answering Challenge},
    author={Siva Reddy and Danqi Chen and Christopher D. Manning},
    year={2018},
    eprint={1808.07042},
    archivePrefix={arXiv},
    primaryClass={cs.CL}
}
"""

_DESCRIPTION = """\
CoQA: A Conversational Question Answering Challenge
"""

_TRAIN_DATA_URL = "https://nlp.stanford.edu/data/coqa/coqa-train-v1.0.json"
_DEV_DATA_URL = "https://nlp.stanford.edu/data/coqa/coqa-dev-v1.0.json"
_HOMEPAGE_URL = "https://stanfordnlp.github.io/coqa/"


class Coqa(tfds.core.GeneratorBasedBuilder):
  """A Conversational Question Answering Challenge."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "source": tfds.features.Text(),
            "story": tfds.features.Text(),
            "questions": tfds.features.Sequence(tfds.features.Text()),
            "answers": tfds.features.Sequence({
                "input_text": tfds.features.Text(),
                "answer_start": np.int32,
                "answer_end": np.int32,
            }),
        }),
        supervised_keys=None,
        homepage=_HOMEPAGE_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    urls_to_download = {"train": _TRAIN_DATA_URL, "test": _DEV_DATA_URL}
    downloaded_files = dl_manager.download_and_extract(urls_to_download)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "filepath": downloaded_files["train"],
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "filepath": downloaded_files["test"],
            },
        ),
    ]

  def _generate_examples(self, filepath):
    """Yields examples."""

    with tf.io.gfile.GFile(filepath, "r") as f:
      data = json.load(f)
      for row in data["data"]:
        questions = [question["input_text"] for question in row["questions"]]
        story = row["story"]
        source = row["source"]
        answers_start = [answer["span_start"] for answer in row["answers"]]
        answers_end = [answer["span_end"] for answer in row["answers"]]
        answers = [answer["input_text"] for answer in row["answers"]]
        yield row["id"], {
            "source": source,
            "story": story,
            "questions": questions,
            "answers": {
                "input_text": answers,
                "answer_start": answers_start,
                "answer_end": answers_end,
            },
        }
